from argparse import ArgumentParser, FileType, Namespace

from .collect_framework import unique_symbol
from .exceptions import UniqueArgsError


def get_args() -> Namespace:
    parser = ArgumentParser(
        description="counting unique characters in a string or file"
    )
    parser.add_argument("--file", type=FileType(), help="file to count")
    parser.add_argument("--string", help="string to count")
    return parser.parse_args()


def get_result(args: Namespace | None = None) -> int:
    args = args or get_args()
    if args.file:
        return unique_symbol(args.file.read())
    if args.string:
        return unique_symbol(args.string)
    raise UniqueArgsError("No arguments")



import io
from argparse import Namespace
from unittest.mock import patch

import pytest

from collection_framework.cli import get_result
from collection_framework.exceptions import UniqueArgsError

STRING_TEXT = "abcd"
FILE_TEXT = "abcdefg\nhi"


@patch("collection_framework.cli.ArgumentParser.parse_args")
@patch("collection_framework.cli.unique_symbol")
class TestCLI():

    def test_exception(self, _, parse_args_mock):
        with pytest.raises(UniqueArgsError) as excinfo:
            parse_args_mock.return_value = Namespace(string=None, file=None)
            get_result()
        assert excinfo.value.args[0] == "No arguments"

    def test_string(self, unique_symbol_mock, parse_args_mock):
        parse_args_mock.return_value = Namespace(string=STRING_TEXT, file=None)
        unique_symbol_mock.return_value = 4
        assert get_result() == 4
        unique_symbol_mock.assert_called_once_with(STRING_TEXT)

    def test_file(self, unique_symbol_mock, parse_args_mock):
        parse_args_mock.return_value = Namespace(
            string=None, file=io.TextIOWrapper(
                io.BytesIO(FILE_TEXT.encode("utf-8"))
            )
        )
        unique_symbol_mock.return_value = 10
        assert get_result() == 10
        unique_symbol_mock.assert_called_once_with(FILE_TEXT)

    def test_both(self, unique_symbol_mock, parse_args_mock):
        parse_args_mock.return_value = Namespace(
            string=STRING_TEXT, file=io.TextIOWrapper(
                io.BytesIO(FILE_TEXT.encode("utf-8"))
            )
        )
        unique_symbol_mock.return_value = 10
        assert get_result() == 10
        unique_symbol_mock.assert_called_once_with(FILE_TEXT)



from collections import Counter
from functools import cache

from .exceptions import UniqueTypeError


@cache
def __unique_count(text: str) -> int:
    return tuple(Counter(text).values()).count(1)


def unique_symbol(text: str) -> int:
    if not isinstance(text, unique_symbol.__annotations__["text"]):
        raise UniqueTypeError(f"'text' must be {unique_symbol.__annotations__['text']}")
    return __unique_count(text)


def unique_symbol_in_list(strings: list) -> list[int]:
    if not isinstance(strings, unique_symbol_in_list.__annotations__["strings"]):
        raise UniqueTypeError(
            f"'strings' must be {unique_symbol_in_list.__annotations__['strings']}"
        )
    return list(map(unique_symbol, strings))



import pytest

from collection_framework.collect_framework import (UniqueTypeError,
                                                    unique_symbol,
                                                    unique_symbol_in_list)

data = {
    unique_symbol: {
        "correct": (("abc", 3), ("", 0), ("abbbccdf", 3)),
        "exception": ((1,), (1.5,), (["a", "b", "c"],)),
    },
    unique_symbol_in_list: {
        "correct": (
            (["abc", "difj", "hi"], [3, 4, 2]),
            ([""], [0]),
            (["abbbccdf"], [3]),
        ),
        "exception": (([1],), ([1.5],), ([["a", "b", "c"]],)),
    },
}


def get_data(data, key):
    argvalues = [(f, *params) for f, value in data.items() for params in value[key]]
    return {
        "argvalues": argvalues,
        "ids": [
            f"{v[0].__name__}({v[1].__repr__()})"
            for v in argvalues
        ],
    }


@pytest.mark.parametrize("f,p,r", **get_data(data, "correct"))
def test_correct(f, p, r):
    assert f(p) == r


@pytest.mark.parametrize("f,p", **get_data(data, "exception"))
def test_exception(f, p):
    with pytest.raises(UniqueTypeError):
        f(p)


def test_cache():
    def work_time(data):
        from timeit import timeit

        return timeit(lambda: unique_symbol(data), number=1)

    data = "a" * 1000000
    assert work_time(data) / work_time(data) > 10
